{% load i18n l10n cms_tags cms_js_tags %}

<div id="cms-top"
     class="cms cms-reset
     {% if not user.is_authenticated %} cms-toolbar-auth{% endif %}
     {% if debug %} cms-toolbar-debug{% endif %}"
     dir="ltr" data-touch-action="none">

    {% block toolbar_top %}
    {% endblock toolbar_top %}

    {# start: toolbar #}
    <div class="cms-toolbar">
        <div class="cms-debug-bar"></div>
        <div class="cms-toolbar-left">
            {% include "cms/toolbar/items/logo.html" %}
            {% if not user.is_authenticated %}
                {% include "cms/toolbar/items/login.html" %}
            {% else %}
                <ul class="cms-toolbar-item cms-toolbar-item-navigation">
                    {% for item in cms_toolbar.get_left_items %}
                        {{ item.render }}
                    {% endfor %}
                    <li class="cms-toolbar-more">
                        <a href="#">
                            <span>{% trans "More" %}...<span class="cms-icon cms-icon-arrow"></span></span>
                        </a>
                        <ul data-touch-action="pan-y"></ul>
                    </li>
                </ul>
            {% endif %}
        </div>

        <div class="cms-toolbar-right">
            {% for item in cms_toolbar.get_right_items %}
                {{ item.render }}
            {% endfor %}
        </div>
    </div>
    {# end: toolbar #}

    {% block toolbar_bottom %}
    {% endblock toolbar_bottom %}
</div>
